"""
Flameingo Connection Layer
==========================

Network connection layer over UDP for sending serializable message objects.
What do I mean by 'serializable'...?  Take a look at 'm_Messages.py'.

TODO: Message data representation should not be a part of the connection layer.
      Seperate to two modules.


The (optional) ACK Mechanism
----------------------------

By default, no ACK is required for your message.
If your message was lost, you can forget about it.
You can force a message to require an ACK, so that the message will be sent
again and again (every RETRANSMISSION_INTERVAL_SECONDS) until an ACK message
is received from the remote host with matching ACK serial number.
Be aware that when enabling ACK, your message may duplicate in the remote peer.

      
Programmer Interface  -  a simplified guide
--------------------

CONNECTION HANDLING

1. You create a FlameingoConnectionLayer instance.

2. Init the layer with a port by calling 'Init'.

3. Call the layer 'Poll' method every once in a while (typically 100ms)
   because it is not threaded.
   
4. The layer always listens for connections.
   To handle this, call 'RegisterNewConnectionHandler' with your handler method.
   Your handler method must except two arguments:
       sender - the layer itself
       host   - the new connected host

5. You may also want to handle errors.
   Call 'RegisterErrorHandler' with your error handler method.
   Your handler method must except two arguments:
       sender - the layer itself
       text   - string description of the error

6. Call 'AsyncConnect' to connect to a remote host.
   This method returns immediatlly.
   When connection is established (which is not promised to happen), you will
   be notified by your connection handler.


Protocol Overview
-----------------

The protocol header struct is 4 bytes long with optional 2 more bytes, if 
message requires ACK (or message is ACK):
    
       1 byte      - protocol signature. always 70 ('F'), for validity check
       1 byte      - protocol version
       1 byte      - flags:
                     0  - ACK_REQUIRED
                     1  - unused
                     2  - unused
                     3  - unused
                     4  - unused
                     5  - unused
                     6  - unused
                     7  - unused
       1 byte      - message type
       2 bytes     - ack serial number.
                     exists only if ACK_REQUIRED == True or message type is
                     ACK.
"""


import struct, sys, select, time
from socket import *


FLAG_ACK_REQUIRED               = 1

PROTOCOL_SIGNATURE              = 70
PROTOCOL_VERSION                = 1


MESSAGE_TYPE_SYN                = 0
MESSAGE_TYPE_ACK                = 1
MESSAGE_TYPE_KEEP_ALIVE         = 2
MESSAGE_TYPE_LAST_RESERVED      = 3


CONNECTION_TIMEOUT_SECONDS      = 3
KEEP_ALIVE_INTERVAL_SECONDS     = 0.6
RETRANSMISSION_INTERVAL_SECONDS  = 1



class ProtocolHeader:
    def __init__(self):
        self.version = 0
        self.flag_ack_required = 0
        self.ack_serial = 0
        self.header_length = 4
        self.message_type = -1
    

class HeaderEncoder:
    
    def Encode(self, header):
        flags = 0 | (header.flag_ack_required & 1)
        retVal = struct.pack('BBBB', 
                             PROTOCOL_SIGNATURE, 
                             PROTOCOL_VERSION, 
                             flags,
                             header.message_type)
        if header.flag_ack_required or header.message_type == MESSAGE_TYPE_ACK:
            retVal += struct.pack('<H', header.ack_serial)
        return retVal


    
class HeaderParser:
        
    def Parse(self, strData):
        if len(strData) < 4:
            print "[debug] received unknown UDP message. (size < 4)"
            return None
        header = ProtocolHeader()
        signature, header.version, flags, header.message_type = struct.unpack('BBBB', strData[:4])
        if signature != PROTOCOL_SIGNATURE:
            print "[debug] received unknown UDP message. (wrong signature)"
            return None
        if header.version != PROTOCOL_VERSION:
            print "[debug] dropping UDP message (version != " + str(PROTOCOL_VERSION) + ")"
            return None
        header.flag_ack_required = flags & 1
        if header.flag_ack_required or header.message_type == MESSAGE_TYPE_ACK:
            if len(strData) < 6:
                print "[debug] header too small to contain ack serial"
                return None
            header.header_length = 6
            header.ack_serial = struct.unpack('<H', strData[4:6])[0]
        return header



class AckRequiredMessage:
    """
    a message waiting for ack
    """
    def __init__(self, addr, raw_data, ack_serial):
        self.addr = addr
        self.raw_data = raw_data
        self.ack_serial = ack_serial
        self.last_retransmition_time = time.time()
    
    

class ErrorSender:

    def __init__(self):
        self.error_handlers = []
        
    def RegisterErrorHandler(self, handlerMethod):
        self.error_handlers.append(handlerMethod)
        
    def NotifyError(self, text):
        for h in self.error_handlers:
            h(self, text)
    

            
class FlameingoConnectionLayer(ErrorSender):
    
    def __init__(self):
        ErrorSender.__init__(self)
        self.socket = None
        self.headerParser = HeaderParser()
        self.headerEncoder = HeaderEncoder()

    def RegisterNewConnectionHandler(self, handlerMethod):
        self.new_connection_handlers.append(handlerMethod)
    
    def Init(self, port=0):
        self.Close()
        self.socket = socket(AF_INET, SOCK_DGRAM)
        try:
            self.socket.bind(('', port))
        except:
            self.NotifyError("can't bind to port " + str(port) + ": " + str(sys.exc_value))
            self.socket = None
            return False
        self.new_connection_handlers = []
        self.connections = {} # from address tuple ('ip_addr', port) to Connection object
        self.ack_serial = 0
        self.waiting_for_ack = {} # from ack serial to AckRequiredMessage
        return True
    
    def Close(self):
        if self.socket != None:
            self.socket.close()
            self.socket = None

    def Poll(self):
        if self.socket == None: return
        while self.PollRead() > 0: pass
        time_now = time.time()
        self._check_connections_timeout_(time_now)
        self._retransmit_ack_required_(time_now)

    def PollRead(self):
        data = None
        addr = None
        try:
            ret = select.select([self.socket], [], [], 0)[0]
            if len(ret) > 0:
                data, addr = self.socket.recvfrom(32000)
        except:
            self.NotifyError('socket error: ' + str(sys.exc_value))
            return 0
        if data != None:
            header = self.headerParser.Parse(data)
            if header != None:    
                self.HandleNewMessage(addr, header, data[header.header_length:])
            return len(data)
        else:
            return 0

    def HandleNewMessage(self, addr, header, data):
        if header.message_type == MESSAGE_TYPE_ACK:
            if header.ack_serial in self.waiting_for_ack:
                ackmsg = self.waiting_for_ack[header.ack_serial]
                if ackmsg.addr == addr:
                    self.waiting_for_ack.pop(header.ack_serial)
            return
        if header.flag_ack_required:
            self._send_ack_(addr, header.ack_serial)
        if addr in self.connections:
            connection = self.connections[addr]
            connection.last_receive_time = time.time()
            if header.message_type > MESSAGE_TYPE_LAST_RESERVED:
                connection.HandleNewMessage(header.message_type, data)
            return
        if header.message_type == MESSAGE_TYPE_SYN:
            self._send_message_(addr, MESSAGE_TYPE_SYN)
            connection = Connection(self, addr)
            self.connections[addr] = connection
            for h in self.new_connection_handlers:
                h(connection)
    
    def _send_message_(self, addr, message_type, data="", flags=0):
        header = ProtocolHeader()
        header.message_type = message_type
        if flags & FLAG_ACK_REQUIRED:
            header.flag_ack_required = 1
            header.ack_serial = self.ack_serial
            data = self.headerEncoder.Encode(header) + data
            ackmsg = AckRequiredMessage(addr, data, self.ack_serial)
            self.waiting_for_ack[self.ack_serial] = ackmsg
            self.ack_serial += 1
        else:
            data = self.headerEncoder.Encode(header) + data
        self._send_raw_data_(addr, data)
        
    def _send_ack_(self, addr, ack_serial):
        header = ProtocolHeader()
        header.message_type = MESSAGE_TYPE_ACK
        header.flag_ack_required = 0
        header.ack_serial = ack_serial
        data = self.headerEncoder.Encode(header)
        self._send_raw_data_(addr, data)

    def _retransmit_ack_required_(self, time_now):
        for msg in self.waiting_for_ack.values():
            if time_now - msg.last_retransmition_time > RETRANSMISSION_INTERVAL_SECONDS:
                msg.last_retransmition_time = time_now
                self._send_raw_data_(msg.addr, msg.raw_data)
        
    def _check_connections_timeout_(self, time_now):
        for connection in self.connections.values():
            if time_now - connection.last_receive_time > CONNECTION_TIMEOUT_SECONDS:
                connection.Close()
            elif time_now - connection.last_send_time > KEEP_ALIVE_INTERVAL_SECONDS:
                self._send_message_(connection.addr, MESSAGE_TYPE_KEEP_ALIVE)
                
    def _send_raw_data_(self, addr, data):
        try:
            self.socket.sendto(data, addr)
        except:
            self.NotifyError("can't send data to " + str(addr) + ": " + str(sys.exc_value))
        if addr in self.connections:
            self.connections[addr].last_send_time = time.time()
    
    def AsyncConnect(self, ip_addr, port):
        self._send_message_( (ip_addr, port), MESSAGE_TYPE_SYN)
        
    def CloseConnection(self, connection):
        if connection.addr in self.connections:
            self.connections.pop(connection.addr)
            keys_to_delete = []
            for ackmsg in self.waiting_for_ack.values():
                if ackmsg.addr == connection.addr:
                    keys_to_delete.append(ackmsg.ack_serial)
            for key in keys_to_delete:
                self.waiting_for_ack.pop(key)
        
    def BreakNAT(self, host_list, port_min, port_max):
        while port_min <= port_max:
            for host in host_list:
                addr = (host, port_min)
                self._send_raw_data_(addr, "la la la")
            port_min += 1

    
    
class Connection:

    def __init__(self, fsocket, addr):
        self.fsocket = fsocket
        self.addr = addr
        self.new_message_handlers = []
        self.connection_closed_handlers = []
        self.last_receive_time = time.time()
        self.last_send_time = 0
        self.closed = False
        
    def Close(self):
        if not self.closed:
            self.closed = True
            self.fsocket.CloseConnection(self)
            for h in self.connection_closed_handlers:
                h(self)
    
    def SendMessage(self, msg_type, data, flags=0):
        if not self.addr in self.fsocket.connections:
            print "[debug] error sending message: this connection is closed (" + str(self.addr) + ")"
        else:
            self.fsocket._send_message_(self.addr, msg_type, data, flags)
    
    def RegisterNewMessageHandler(self, handleMethod):
        self.new_message_handlers.append(handleMethod)

    def RegisterConnectionClosedHandler(self, handleMethod):
        self.connection_closed_handlers.append(handleMethod)
        
    def HandleNewMessage(self, msg_type, data):
        for h in self.new_message_handlers:
            h(self, msg_type, data)



