"""
FlameingoIPC
============

A message delivery mechanism over the unreliable FlameingoConnectionLayer.

A message is an object defined in 'm_Messages' which has a 'serializer'.
Every message ***class*** has an integer type.

When you accept a new connection with your connection handler method, you
may want to handle messages it receives.
You bind a specific message type to a specific message handler by calling
the method 'RegisterObjectReceivedHandler' with a type and your handler 
method. Your handler method must except two arguments:
    sender - the RemoteHost object you registered
    msg    - message object. 
   
To send a message, call the remote host method 'SendObject'.
This method expects two arguments:
    msg               - a message object.
    flags             - for the connection layer.

"""


from m_FlameingoConnectionLayer import FlameingoConnectionLayer, ErrorSender, FLAG_ACK_REQUIRED
from m_Messages import all_message_types
import time, socket


class FlameingoIPC(ErrorSender):

    def __init__(self):
        ErrorSender.__init__(self)
        self.connection_layer = FlameingoConnectionLayer()
        self.initialized = False

    def Init(self, port=0):
        if self.connection_layer.Init(port) == False:
            return False
        self.connection_layer.RegisterErrorHandler(self._handle_connection_layer_error_)
        self.connection_layer.RegisterNewConnectionHandler(self._handle_new_connection_)
        self.new_connection_handlers = []
        self.waiting_connections = {} # used by CreateNewConnection. from addr to RemoteHost or None
        self.initialized = True
        return True
    
    def Close(self):
        if self.initialized:
            self.connection_layer.Close()
            self.initialized = False
    
    def Poll(self):
        self.connection_layer.Poll()
        
    def RegisterNewConnectionHandler(self, handlerMethod):
        self.new_connection_handlers.append(handlerMethod)
        
    def CreateNewConnection(self, ip_addr, port):
        try:
            ip_addr = socket.gethostbyname(ip_addr)
        except:
            self.NotifyError("failed in name resolution")
            return None
        addr = (ip_addr, port)
        self.waiting_connections[addr] = None
        self.connection_layer.AsyncConnect(ip_addr, port)
        timeout = 2 # seconds
        while timeout > 0:
            self.Poll()
            if self.waiting_connections[addr] != None:
                return self.waiting_connections.pop(addr)
            time.sleep(0.1)
            timeout -= 0.1
        return None

    def _handle_connection_layer_error_(self, sender, text):
        self.NotifyError(text)

    def _handle_new_connection_(self, connection):
        host = RemoteHost(connection)
        if connection.addr in self.waiting_connections:
            self.waiting_connections[connection.addr] = host
            return
        for h in self.new_connection_handlers:
            h(host)
    
    def BreakNAT(self, host_list, port_min, port_max):
        self.connection_layer.BreakNAT(host_list, port_min, port_max)

    
        
class RemoteHost:

    def __init__(self, connection):
        self.connection = connection
        self.addr = connection.addr
        self.object_received_handlers = {}
        self.connection_closed_handlers = []
        self.connection.RegisterNewMessageHandler(self.HandleNewMessage)
        self.connection.RegisterConnectionClosedHandler(self.HandleConnectionClosed)
        self.closed = False
        
    def Close(self):
        if not self.closed:
            self.closed = True
            self.connection.Close()
    
    def SendObject(self, obj, flags=0):
        data = obj.serializer.Serialize(obj)
        self.connection.SendMessage(obj.type, data, flags)
    
    def RegisterNewMessageHandler(self, handleMethod):
        self.new_message_handlers.append(handleMethod)
    
    def RegisterObjectReceivedHandler(self, objType, handleMethod):
        if self.object_received_handlers.has_key(objType):
            self.object_received_handlers[objType].append(handleMethod)
        else:
            self.object_received_handlers[objType] = [handleMethod]

    def RegisterConnectionClosedHandler(self, handleMethod):
        self.connection_closed_handlers.append(handleMethod)
        
    def HandleConnectionClosed(self, sender):
        for h in self.connection_closed_handlers: h(self)
        
    def HandleNewMessage(self, sender, msg_type, data):
        if not msg_type in all_message_types:
            print "[debug] received message of unknown type"
            return
        if msg_type in self.object_received_handlers:
            msg = all_message_types[msg_type]()
            msg.serializer.Deserialize(msg, data)
            for h in self.object_received_handlers[msg_type]:
                h(self, msg)

