"""
This is a module for class serializiation, which means taking an object and 
returning a buffer, and vise versa.

If all members are primitives, the resulting buffer gives a struct, like in C.
It also supports 'list' and other serializeable classes.

The important advantages are:
    o  You have controll of the struct order (structure), so you can specify
       your protocol exactlly.
    o  The result is smaller than in 'pickle'

    
Here is a simple example:
========================

    class PlayerInfo:
        def __init__(self, player_name=None, lives=None):
            self.player_name = player_name     # null-terminating string
            self.lives = lives                 # unsigned byte

        # this is what you need to add:            
        serializer = ClassSerializer()
        serializer.DefineAttribute("player_name",   Type_str)
        serializer.DefineAttribute("lives",         Type_u8)

    # now let's see it in action:
    
    i = PlayerInfo("Momo", 7)
    s = PlayerInfo.serializer.Serialize(i)
    
    # now 's' is a struct! like in C!
    
    j = PlayerInfo()
    PlayerInfo.serializer.Deserialize(j)
    
    # now 'j' is equal to 'i'. ta-da!
    

More complex example combining class in a class and list of classes
is at the end of this file.


@Author: Gal Koren
"""


import struct


# first we define 'types'.
# a 'type' is an object (or class, or whatever) with two methods:
#     o  serialize - turns a value to binary string
#     o  deserialize - turns a binary string to a value, and (this is the
#                      important part) returns the number of bytes read
#                      from the string and the value itself, as a tuple.

###############################################################################
# primitive types
###############################################################################

class Type_s8:
    
    def static_serialize(pValue):
        return struct.pack('b', pValue)
    
    def static_deserialize(pValue):
        return (1, struct.unpack('b', pValue[:1])[0])
    
    serialize = staticmethod(static_serialize)
    deserialize = staticmethod(static_deserialize)


class Type_u8:
    
    def static_serialize(pValue):
        return struct.pack('B', pValue)
    
    def static_deserialize(pValue):
        return (1, struct.unpack('B', pValue[:1])[0])
    
    serialize = staticmethod(static_serialize)
    deserialize = staticmethod(static_deserialize)

    
class Type_s16:
    
    def static_serialize( pValue):
        return struct.pack('<h', pValue)
    
    def static_deserialize(pValue):
        return (2, struct.unpack('<h', pValue[:2])[0])
    
    serialize = staticmethod(static_serialize)
    deserialize = staticmethod(static_deserialize)


class Type_u16:
    
    def static_serialize( pValue):
        return struct.pack('<H', pValue)
    
    def static_deserialize(pValue):
        return (2, struct.unpack('<H', pValue[:2])[0])
    
    serialize = staticmethod(static_serialize)
    deserialize = staticmethod(static_deserialize)


class Type_s32:
    
    def static_serialize( pValue):
        return struct.pack('<l', pValue)
    
    def static_deserialize(pValue):
        return (4, struct.unpack('<l', pValue[:4])[0])
    
    serialize = staticmethod(static_serialize)
    deserialize = staticmethod(static_deserialize)


class Type_u32:
    
    def static_serialize( pValue):
        return struct.pack('<L', pValue)
    
    def static_deserialize(pValue):
        return (4, struct.unpack('<L', pValue[:4])[0])
    
    serialize = staticmethod(static_serialize)
    deserialize = staticmethod(static_deserialize)
    

class Type_f32:
    
    def static_serialize( pValue):
        return struct.pack('<f', pValue)
    
    def static_deserialize(pValue):
        return (4, struct.unpack('<f', pValue[:4])[0])
    
    serialize = staticmethod(static_serialize)
    deserialize = staticmethod(static_deserialize)


class Type_f64:
    
    def static_serialize( pValue):
        return struct.pack('<d', pValue)
    
    def static_deserialize(pValue):
        return (8, struct.unpack('<d', pValue[:8])[0])
    
    serialize = staticmethod(static_serialize)
    deserialize = staticmethod(static_deserialize)
    
        
class Type_str:
    """
    The result is a null terminating string.
    """
    
    def static_serialize( pValue):
        return pValue + '\0'
    
    def static_deserialize(pValue):
        index = pValue.find('\0')
        if index == -1: index = len(pValue)
        return (index + 1, pValue[:index])
    
    serialize = staticmethod(static_serialize)
    deserialize = staticmethod(static_deserialize)
    

###############################################################################
# more complex types
###############################################################################


# the follosing classes need to be instantiated with a specific other type
# before usage. see example.


class Type_list:
    """
    The result is 2 bytes specifing the number of elements, and then
    each element serialized according to it's type.
    """
            
    def __init__(self, itemType):
        self.itemType = itemType
        
    def serialize(self, pValue):
        retVal = struct.pack('<H', len(pValue))
        for item in pValue:
            retVal += self.itemType.serialize(item)
        return retVal
    
    def deserialize(self, pValue):
        count = struct.unpack('<H', pValue[:2])[0]
        pValue = pValue[2:]
        size = 2
        retList = []
        while count > 0:
            n, item = self.itemType.deserialize(pValue)
            size += n
            pValue = pValue[n:]
            retList.append(item)
            count -= 1
        return (size, retList)
    

class Type_class:
    """
    Use this type to include a serialized class in another serialized class.
    The child class must have a 'serializer' static member.
    """
    
    def __init__(self, pClass):
        """
        pClass must have a static 'serializer' member
        """
        self.objClass = pClass
        
    def serialize(self, obj):
        return self.objClass.serializer.Serialize(obj)
    
    def deserialize(self, pString):
        obj = self.objClass()
        count = self.objClass.serializer.Deserialize(obj, pString)
        return (count,  obj)

                
###############################################################################
# The class which is everything
###############################################################################

class ClassSerializer:
    """
    a serializer for a specific class.
    """
    
    def __init__(self):
        ### list of (attr_name, Type_x) tuples
        self.attrList = []
        
    def DefineAttribute(self, attrName, attrType):
        self.attrList.append( (attrName, attrType) )
        
    def Serialize(self, obj):
        """
        returns a string.
        """
        retVal = ""
        for attrName, typeClass in self.attrList:
            retVal += typeClass.serialize(obj.__dict__[attrName])
        return retVal
        
    def Deserialize(self, obj, pString):
        """
        update attributes of obj with data from pString.
        returns the number of bytes read from the string.
        """
        count = 0
        for attrName, typeClass in self.attrList:
            bytes, val = typeClass.deserialize(pString)
            obj.__dict__[attrName] = val
            pString = pString[bytes:]
            count += bytes
        return count
        

    
        
# A complex example
# =================
#
#
# class PlayerInfo:
#     def __init__(self, player_name=None, lives=None):
#         self.player_name = player_name
#         self.lives = lives
#         
#     serializer = ClassSerializer()
#     serializer.DefineAttribute("player_name", Type_str)
#     serializer.DefineAttribute("lives", Type_u8)
# 
# 
# class Test:
#     def __init__(self):
#         self.level_filename = ""
#         self.state = 0
#         self.count = 0
#         self.players = []
#         self.numbers = []
#         self.x = 0
#         self.y = 0
#         
#     serializer = ClassSerializer()
#     serializer.DefineAttribute("players", Type_list(Type_class(PlayerInfo)))
#     serializer.DefineAttribute("level_filename", Type_str)
#     serializer.DefineAttribute("state", Type_u8)
#     serializer.DefineAttribute("count", Type_u16)
#     serializer.DefineAttribute("numbers", Type_list(Type_u8))
#     serializer.DefineAttribute("x", Type_f32)
#     serializer.DefineAttribute("y", Type_f32)
#         
# 
# t = Test()
# 
# t.state = 20
# t.count = 10
# t.x = 100.15
# t.y = 31.231
# t.numbers = [1, 3, 5, 7, 9]
# t.players = [PlayerInfo("gal koren", 7), PlayerInfo("chiko", 5)]
# t.level_filename = "gasolina.lev"
# 
# s = Test.serializer.Serialize(t)
# print "Serialized string len:", len(s)
# print ""
# 
# t2 = Test()
# 
# print "before"
# print ""
# print "state: ", t2.state
# print "count: ", t2.count
# print "x: ", t2.x
# print "y: ", t2.y
# print "numbers: ", t2.numbers
# for p in t2.players:
#     print "player: { name:", p.player_name, ", lives:", p.lives, "}"
# print "level_filename: ", t2.level_filename
# print ""
# 
# Test.serializer.Deserialize(t2, s)
# 
# print "after"
# print ""
# print "state: ", t2.state
# print "count: ", t2.count
# print "x: ", t2.x
# print "y: ", t2.y
# print "numbers: ", t2.numbers
# for p in t2.players:
#     print "player: { name:", p.player_name, ", lives:", p.lives, "}"
# print "level_filename: ", t2.level_filename
# print ""
