"""
FlameingoSocket Layer

This is the connection layer of the FLAMEingo protocol.
It forms a connection between peers over UDP.

The protocol header struct is 4 bytes long with optional 2 more bytes, if ack
is required:
    
       1 byte      - protocol signature. always 70 ('F'), for validity check
       1 byte      - protocol version
       1 byte      - flags:
                     0  - ACK_REQUIRED
                     1  - unused
                     2  - unused
                     3  - unused
                     4  - unused
                     5  - unused
                     6  - unused
                     7  - unused
       1 byte      - message type
       2 bytes     - ack serial number.
                     exists only if ACK_REQUIRED == True or message type is
                     ACK.
"""


import struct, sys, select, time
from socket import *
from m_Messages import messageTypeMap


PROTOCOL_SIGNATURE              = 70
PROTOCOL_VERSION                = 1


MESSAGE_TYPE_SYN                = 0
MESSAGE_TYPE_ACK                = 1
MESSAGE_TYPE_KEEP_ALIVE         = 2
MESSAGE_TYPE_LAST_RESERVED      = 3


CONNECTION_TIMEOUT_SECONDS      = 3
KEEP_ALIVE_INTERVAL_SECONDS     = 0.4
RETRANSMISION_INTERVAL_SECONDS  = 1

class ProtocolHeader:
    def __init__(self):
        self.version = 0
        self.flag_ack_required = 0
        self.ack_serial = 0
        self.header_length = 4
        self.message_type = -1
    

class HeaderEncoder:
    
    def Encode(self, header):
        flags = 0 | (header.flag_ack_required & 1)
        retVal = struct.pack('BBBB', 
                             PROTOCOL_SIGNATURE, 
                             PROTOCOL_VERSION, 
                             flags,
                             header.message_type)
        if header.flag_ack_required or header.message_type == MESSAGE_TYPE_ACK:
            retVal += struct.pack('<H', header.ack_serial)
        return retVal


    
class HeaderParser:
        
    def Parse(self, strData):
        if len(strData) < 4:
            print "[debug] received unknown UDP message. (size < 4)"
            return None
        header = ProtocolHeader()
        signature, header.version, flags, header.message_type = struct.unpack('BBBB', strData[:4])
        if signature != PROTOCOL_SIGNATURE:
            print "[debug] received unknown UDP message. (wrong signature)"
            return None
        if header.version != PROTOCOL_VERSION:
            print "[debug] dropping UDP message (version != " + str(PROTOCOL_VERSION) + ")"
            return None
        header.flag_ack_required = flags & 1
        if header.flag_ack_required or header.message_type == MESSAGE_TYPE_ACK:
            if len(strData) < 6:
                print "[debug] header too small to contain ack serial"
                return None
            header.header_length = 6
            header.ack_serial = struct.unpack('<H', strData[4:6])[0]
        return header



class AckRequiredMessage:
    """
    a message waiting for ack
    """
    def __init__(self, addr, raw_data, ack_serial):
        self.addr = addr
        self.raw_data = raw_data
        self.ack_serial = ack_serial
        self.last_retransmition_time = time.time()
    
    

class ErrorSender:

    def __init__(self):
        self.error_handlers = []
        
    def RegisterErrorHandler(self, handlerMethod):
        self.error_handlers.append(handlerMethod)
        
    def NotifyError(self, text):
        for h in self.error_handlers:
            h(self, text)
    

            
class FlameingoSocket(ErrorSender):
    
    def __init__(self):
        ErrorSender.__init__(self)
        self.new_connection_handlers = []
        self.remote_hosts = {} # from address tuple ('ip_addr', port) to RemoteHost
        self.socket = socket(AF_INET, SOCK_DGRAM)
        self.headerParser = HeaderParser()
        self.headerEncoder = HeaderEncoder()
        self.ack_serial = 0
        self.waiting_for_ack = {} # from ack serial to AckRequiredMessage

    def RegisterNewConnectionHandler(self, handlerMethod):
        self.new_connection_handlers.append(handlerMethod)
    
    def Bind(self, port=0):
        try:
            self.socket.bind(('', port))
        except:
            self.NotifyError("can't bind to port " + str(port) + ": " + str(sys.exc_value))
            return False
        return True
    
    def Poll(self):
        while self.PollRead() > 0: pass
        time_now = time.time()
        self._check_connections_timeout_(time_now)
        self._retransmit_ack_required_(time_now)

    def PollRead(self):
        data = None
        addr = None
        try:
            ret = select.select([self.socket], [], [], 0)[0]
            if len(ret) > 0:
                data, addr = self.socket.recvfrom(32000)
        except:
            self.NotifyError('socket error: ' + str(sys.exc_value))
            return 0
        if data != None:
            header = self.headerParser.Parse(data)
            if header != None:    
                self.HandleNewMessage(addr, header, data[header.header_length:])
            return len(data)
        else:
            return 0

    def HandleNewMessage(self, addr, header, data):
        if header.message_type == MESSAGE_TYPE_ACK:
            print "===> received ack", header.ack_serial
            if header.ack_serial in self.waiting_for_ack:
                ackmsg = self.waiting_for_ack[header.ack_serial]
                if ackmsg.addr == addr:
                    self.waiting_for_ack.pop(header.ack_serial)
            return
        if header.flag_ack_required:
            self._send_ack_(addr, header.ack_serial)
        if addr in self.remote_hosts:
            host = self.remote_hosts[addr]
            host.last_receive_time = time.time()
            if header.message_type in messageTypeMap:
                objClass = messageTypeMap[header.message_type]
                msg = objClass()
                objClass.serializer.Deserialize(msg, data)
                host.HandleNewMessage(msg)
            elif header.message_type > MESSAGE_TYPE_LAST_RESERVED:
                print "[debug] received message with unknown type: " + str(header.message_type)
            return
        if header.message_type == MESSAGE_TYPE_SYN:
            self._send_message_(addr, MESSAGE_TYPE_SYN)
            host = RemoteHost(self, addr)
            self.remote_hosts[addr] = host
            for h in self.new_connection_handlers:
                h(host)   
    
    def _send_message_(self, addr, message_type, data="", require_ack=0):
        header = ProtocolHeader()
        header.message_type = message_type
        if require_ack:
            print "===> sending require ack", self.ack_serial
            header.flag_ack_required = require_ack
            header.ack_serial = self.ack_serial
            data = self.headerEncoder.Encode(header) + data
            ackmsg = AckRequiredMessage(addr, data, self.ack_serial)
            self.waiting_for_ack[self.ack_serial] = ackmsg
            self.ack_serial += 1
        else:
            data = self.headerEncoder.Encode(header) + data
        self._send_raw_data_(addr, data)
        
    def _send_ack_(self, addr, ack_serial):
        print "===> sending ack",ack_serial
        header = ProtocolHeader()
        header.message_type = MESSAGE_TYPE_ACK
        header.flag_ack_required = 0
        header.ack_serial = ack_serial
        data = self.headerEncoder.Encode(header)
        self._send_raw_data_(addr, data)

    def _retransmit_ack_required_(self, time_now):
        for msg in self.waiting_for_ack.values():
            if time_now - msg.last_retransmition_time > RETRANSMISION_INTERVAL_SECONDS:
                msg.last_retransmition_time = time_now
                print "===> retransmition. ack_serial:", msg.ack_serial
                self._send_raw_data_(msg.addr, msg.raw_data)
        
    def _check_connections_timeout_(self, time_now):
        for host in self.remote_hosts.values():
            if time_now - host.last_receive_time > CONNECTION_TIMEOUT_SECONDS:
                host.Close()
            elif time_now - host.last_send_time > KEEP_ALIVE_INTERVAL_SECONDS:
                self._send_message_(host.addr, MESSAGE_TYPE_KEEP_ALIVE)
                
    def _send_raw_data_(self, addr, data):
        try:
            self.socket.sendto(data, addr)
        except:
            self.NotifyError("can't send data to " + str(addr) + ": " + str(sys.exc_value))
        if addr in self.remote_hosts:
            self.remote_hosts[addr].last_send_time = time.time()
    
    def ConnectToHost(self, ip_addr, port):
        self._send_message_( (ip_addr, port), MESSAGE_TYPE_SYN)
        
    def CloseRemoteHost(self, host):
        if host.addr in self.remote_hosts:
            self.remote_hosts.pop(host.addr)
            keys_to_delete = []
            for ackmsg in self.waiting_for_ack.values():
                if ackmsg.addr == host.addr:
                    keys_to_delete.append(ackmsg.ack_serial)
            for key in keys_to_delete:
                self.waiting_for_ack.pop(key)
        

    
    
class RemoteHost:

    def __init__(self, fsocket, addr):
        self.fsocket = fsocket
        self.addr = addr
        self.object_received_handlers = {}
        self.connection_closed_handlers = []
        self.last_receive_time = time.time()
        self.last_send_time = 0
        
    def Close(self):
        self.fsocket.CloseRemoteHost(self)
        for h in self.connection_closed_handlers:
            h(self)
    
    def SendMessage(self, msg, require_ack=0):
        if not self.addr in self.fsocket.remote_hosts:
            print "[debug] trying to send message to disconnected host (" + str(self.addr) + ")"
            return
        self.fsocket._send_message_(self.addr, msg.type, msg.serializer.Serialize(msg), require_ack)
    
    def RegisterObjectReceivedHandler(self, objType, handleMethod):
        if self.object_received_handlers.has_key(objType):
            self.object_received_handlers[objType].append(handleMethod)
        else:
            self.object_received_handlers[objType] = [handleMethod]

    def RegisterConnectionClosedHandler(self, handleMethod):
        self.connection_closed_handlers.append(handleMethod)
        
    def HandleNewMessage(self, msg):
        if self.object_received_handlers.has_key(msg.type):
            for h in self.object_received_handlers[msg.type]:
                h(self, msg)



