"""
The model part of the scene.

The heart of the model is the Entity and Level classes.
"""

import time # slowing down the thread


SIDE_TOP     = 0
SIDE_BOTTOM  = 1
SIDE_LEFT    = 2
SIDE_RIGHT   = 3


class Outlines:
    """
    Define the outlines of an Entity by defining four lists of
    points: left, right, top and bottom.
    @ivar left: list of points defining the left outline.
    @ivar right: list of points defining the right outline.
    @ivar top: list of points defining the top outline.
    @ivar bottom: list of points defining the bottom outline.
    """
    
    def __init__(self):
        self.left = []
        self.right = []
        self.top = []
        self.bottom = []
    
    
class EntityAnimation:
    
    def __init__(self, num_of_frames, delay):
        """
        Constructs a new EntityAnimation object.
        @param frames: list of EntityFrame objects.
        @param delay: delay in milliseconds between
        the frames.
        """
        self.num_of_frames = num_of_frames
        self.delay = delay
        self.time = 0
        self.frame_index = 0
        self.cycles = 0
        
    def Advance(self, time):
        """
        Advances the animation.
        @param time: time in milliseconds to advance.
        """
        self.time += time
        while self.time > self.delay:
            self.time -= self.delay
            self.frame_index += 1
            if self.frame_index >= self.num_of_frames:
                self.frame_index = 0
                self.cycles += 1



class IEntityObserver:
    """
    An interface for observer of an Entity. Observers can be
    attached to Entity objects to be notified with callbacks.
    """
    
    def NotifyCollisionWithEntity(self, entity_id1, entity_id2):
        """
        A callback notifying about collision of the observed entity
        with another entity.
        @param entity1: the observerd entity.
        @param entity2: the other entity.
        """
        pass
        
    def NotifyCollisionWithLand(self, entity_id, x, y, side):
        """
        A callback called when the physics engine detects a collision
        with the land.
        @param x: x position of the collision.
        @param y: y position of the collision.
        @param side: either SIDE_TOP, SIDE_BOTTOM,
                     SIDE_LEFT or SIDE_RIGHT.
        """
        pass
            
    def NotifyExplosion(self, entity_id, vx, vy, power):
        pass


class Entity:
    """
    A physical object in the scene with position and speed.
    
    Entity has a collection of EntityAnimation objects which define
    it's outlines and rectangle in the model.
    All subclasses of this class must overload the Clone() method.
    
    @ivar type: the type of the entity (integer)
    @ivar x: x position in the level
    @ivar y: y position in the level
    @ivar vx: x velocity
    @ivar vy: y velocity
    @ivar level: the Level object where this entity contained.
    """

    def __init__(self, res=None, level=None, prototype=None):
        """
        Constructs a new Entity object.
        
        There are two methods to create an entity:
        1. Using an EntityResource object.
        2. Using another Entity object as a prototype for cloning.
        @param res: an EntityResource object.
        @param level: a Level object.
        @param prototype: an Entity object to clone this object from.
        """

        self.observers = []
        self.anims = []
        self.state = 0
        self.states = []
        self.x = 0
        self.y = 0
        self.vx = 0
        self.vy = 0
        self.id = -1
        self._physics_cache = None

        if res != None:
            self.type = res.type
            self.center = res.center
            self.mass = res.mass
            self.level = level
            self.elasticity = res.elasticity
            self.gravity = res.gravity
            self.gravity_saturation = res.gravity_saturation
            self.outlines = res.outlines
            self.rect = res.rect
            for a in res.animations:
                self.anims.append(EntityAnimation(len(a.frames), a.delay))
            self.states = res.states[:]
                
        elif prototype != None:
            self.x = prototype.x
            self.y = prototype.y
            self.vx = prototype.vx
            self.vy = prototype.vy
            self.mass = prototype.mass
            self.elasticity = prototype.elasticity
            self.gravity = prototype.gravity
            self.gravity_saturation = prototype.gravity_saturation
            self.type = prototype.type
            self.level = prototype.level
            self.rect = prototype.rect
            self.center = prototype.center
            self.outlines = prototype.outlines
            for a in prototype.anims:
                self.anims.append(EntityAnimation(a.num_of_frames, a.delay))
            self.states = prototype.states[:]
        
        else: 
            raise Exception("not given a res param nor a prototype param")
        
        self.active_anim_indexes = self.states[self.state]


    def Clone(self):
        """
        Constructs a new Entity object using this object as a prototype.
        @return: a clone of this object.
        """
        return Entity(prototype=self)

    def AttachObserver(self, observer):
        """
        Attaches an observer to be notified of collisions.
        @param observer: IEntityObserver object.
        """
        self.observers.append(observer)
        
    def GetCenter(self):
        return self.center
        
    def Advance(self, time):
        """
        Advances the entity by time milliseconds.
        
        The default implementation is advancing the active animations.
        """
        self.active_anim_indexes = self.states[self.state]
        for i in self.active_anim_indexes:
            self.anims[i].Advance(time)
        
    def CollideWithEntity(self, entity):
        """
        A callback called when the collision detector detects a collision
        with another entity.
        @param entity: the entity collided.
        """
        for o in self.observers:
            o.NotifyCollisionWithEntity(self.id, entity.id)

    def CollideWithLand(self, x, y, side):
        """
        A callback called when the physics engine detects a collision
        with the land.
        @param x: x position of the collision.
        @param y: y position of the collision.
        @param side: either SIDE_TOP, SIDE_BOTTOM,
                     SIDE_LEFT or SIDE_RIGHT.
        """
        for o in self.observers:
            o.NotifyCollisionWithLand(self.id, x, y, side)

    def NotifyExplosion(self, vx, vy, power):
        for o in self.observers:
            o.NotifyExplosion(self.id, vx, vy, power)
        


class EntityFactory:
    """
    A factory of Entity objects.
    
    This class contains Entity objects which it can use as a prototype
    to clone new entities.
    """
    
    def __init__(self):
        self.prototypes = {} # map from type integer to Entity object

    def AddPrototype(self, prototype):
        """
        Add a prototype to the factory.
        @param prototype: an Entity subclass object.
        """
        self.prototypes[prototype.type] = prototype
        
    def Create(self, type):
        """
        Clones a new entity from the prototype which matches the given
        type.
        """
        entity = self.prototypes[type].Clone()
        return entity


  
class ILevelObserver:
    """
    An interface with callback about land removal.
    """
    
    def NotifyLandRemoval(self, level, x, y, shape):
        """
        Callback which notifies about land removal.
        @param level: the Level object.
        @param x: removal x pos.
        @param y: removal y pos.
        @param shape: a Shape object.
        """
        pass


class Level:
    """
    This class holds all the Entity objects in the scene, and the land
    information.
    
    @ivar entities: a map from a type (integer) to a list of entities of
                    that type.
    @ivar collidetector: a CollisionDetector object.
    @ivar factory: an EntityFactory object used to create new entities.
    """

    ## a point in the land_map may be one of:
    LAND_NONE    = 0
    LAND_SAND    = 255
    LAND_ROCK    = 1
    
    def __init__(self, land_map, width, height):
        """
        Constructs a new level.
        @param land_map: a 2D array where the value at each point
                         may be:
                         1. Level.LAND_SAND
                         2. Level.LAND_NONE
                         3. Level.LAND_ROCK
        @param width: the level's width
        @param height: the level's height
        """
        self._land_map = land_map
        self._width = width
        self._height = height
        self.entities = {}
        self._all_entities = {} # from id to entity object
        self.id_counter = 0
        self.collidetector = CollisionDetector()
        self.factory = EntityFactory()
        self.observers = []
    
    def GetEntities(self, type=None):
        if type != None:
            if self.entities.has_key(type):
                return self.entities[type]
            else:
                return []
        else:
            return self._all_entities.values()

    def AttachObserver(self, o):
        self.observers.append(o)
    
    def GetWidth(self):
        return self._width
    
    def GetHeight(self):
        return self._height
    
    def GetLandAt(self, x, y):
        if x >= 0 and x < self._width and \
           y >= 0 and y < self._height:
            return self._land_map[int(x)][int(y)]
        else:
            return Level.LAND_NONE
        
    def RemoveLand(self, x, y, shape):
        """
        Removes land from the level.
        @param x: x position of the shape to be removed
        @param y: y position of the shape to be removed
        @param shape: a Shape object of the land to be removed.
        """
        # caching for performance (makes sence only in python)
        LAND_NONE, LAND_SAND = Level.LAND_NONE, Level.LAND_SAND
        w, h  = self._width - 1, self._height - 1
        land_map = self._land_map

        for (y0, x0, x1) in shape.hlines:
            y0 += y
            if y0 < 0 or y0 > h: continue
            x0 = max(0, x + x0)
            x1 = min(w, x + x1)
            while x0 < x1:
                if land_map[x0][y0] == LAND_SAND: land_map[x0][y0] = LAND_NONE
                x0 += 1
            time.sleep(0)
        for o in self.observers:
            o.NotifyLandRemoval(self, x, y, shape)
            
    def AddEntity(self, entity):
        """
        add an entity to the level.
        """
        if entity.id == -1:
            entity.id = self.id_counter
            self.id_counter += 1
        entity.level = self
        self._all_entities[entity.id] = entity
        if self.entities.has_key(entity.type):
            self.entities[entity.type].append(entity)
        else:
            self.entities[entity.type] = [entity]
    
    def RemoveEntity(self, entity):
        """
        remove an entity from the level
        """
        if entity.id in self._all_entities:
            self._all_entities.pop(entity.id)
        if self.entities.has_key(entity.type):
            if entity in self.entities[entity.type]:
                self.entities[entity.type].remove(entity)

        

class CollisionDetector:
    """
    Detects collision between entities.
    
    For performance issue, we wouldn't like to detect collisions between all
    of the entities in the level, but only between types of entities.
    (e.g. between birds and bazooka's)
    This class provides the mechanism to define which types should be detected
    to collide against which other types.
    """
    def __init__(self):
        self._type_map = {}
    
    def DefineCollisionCheck(self, type1, type2):
        """
        defines that collisions should be checked between type1 and type2.
        
        when collision is detected between an entity e1 of type1 and an entity 
        e2 of type2, the method e1.CollideWithEntity(e2) shall be invoked.
        """
        if self._type_map.has_key(type1):
            self._type_map[type1].append(type2)
        else:
            self._type_map[type1] = [type2]
            
    def DetectEntity(self, entity):
        if not self._type_map.has_key(entity.type): return
        rect1 = entity.rect.move(entity.x, entity.y)
        for t2 in self._type_map[entity.type]:
            for entity2 in entity.level.GetEntities(t2):
                rect2 = entity2.rect.move(entity2.x, entity2.y)
                if rect1.colliderect(rect2):
                    entity.CollideWithEntity(entity2)



class _PhysicsCache:
    def __init__(self):
        self.last_bottom_collision_point = None



class PhysicsEngine:

    def __init__(self, level):
        self.level = level
        self.level.AttachObserver(self)
        self.entity_ignore_list = []  # id's of entities that won't be advanced in the method 'Advance'
    
    def NotifyLandRemoval(self, level, x, y, shape):
        ##
        ## clear cache if needed.
        ##
        for entity in level.GetEntities():
            if entity._physics_cache != None:
                p = entity._physics_cache.last_bottom_collision_point
                if p != None and level.GetLandAt(p[0], p[1]) == Level.LAND_NONE:
                    entity._physics_cache.last_bottom_collision_point = None

    def Advance(self, time):
        for entity in self.level.GetEntities():
            if entity.id not in self.entity_ignore_list:
                self.AdvanceEntity(entity, time)
                
    def AdvanceEntity(self, entity, time):
        if entity._physics_cache == None:
            entity._physics_cache = _PhysicsCache()
        if entity.gravity != 0:
            if entity.vy < 0:
                entity._physics_cache.last_bottom_collision_point = None
            if entity._physics_cache.last_bottom_collision_point == None:
                entity.vy += float(entity.gravity * time) / 1000
                if entity.vy > entity.gravity_saturation:
                    entity.vy = entity.gravity_saturation
        # move it
        self._move_entity(entity, time)
        entity.Advance(time)

    def _move_entity(self, entity, time):
        _dx = float(entity.vx) * time / 1000.0
        _dy = float(entity.vy) * time / 1000.0
        dx = int(entity.x + _dx) - int(entity.x)
        dy = int(entity.y + _dy) - int(entity.y)
        steps = max([abs(dx), abs(dy)])
        if steps == 0:
            entity.x += _dx
            entity.y += _dy
            self.level.collidetector.DetectEntity(entity)
            return
        # construct a collection of outlines to check for collisions
        outline_x = None
        outline_y = None
        if dx < 0: outline_x = entity.outlines.left
        elif dx > 0: outline_x = entity.outlines.right
        if dy < 0: outline_y = entity.outlines.top
        elif dy > 0: outline_y = entity.outlines.bottom
        ##
        ## now comes the real work!
        ## because the object may jump over more than one pixel
        ## for each direction in this round, it may not detect
        ## collisions with land in between.
        ## that's why we're going to divide the movement to
        ## jumps of maximum 1 pixel for each direction
        ##
        _dx = _dx / steps
        _dy = _dy / steps
        while steps:
            steps -= 1
            _new_x = entity.x + _dx
            _new_y = entity.y + _dy
            if _new_x < 0: 
                entity.CollideWithLand(-1, -1, SIDE_LEFT)
                dx = _dx = 0
                entity.vx = 0
            elif _new_x >= self.level.GetWidth():
                entity.CollideWithLand(-1, -1, SIDE_RIGHT)
                dx = _dx = 0
                entity.vx = 0
            elif _new_y < 0: 
                entity.CollideWithLand(-1, -1, SIDE_TOP)
                dy = _dy = 0
                entity.vy = 0
            elif _new_y >= self.level.GetHeight():
                entity.vy = 0
                entity.CollideWithLand(-1, -1, SIDE_BOTTOM)
                dy = _dy = 0
            if dx != 0:            
                for pos in outline_x:
                    if self.level.GetLandAt(_new_x + pos[0],
                                            _new_y + pos[1]) != Level.LAND_NONE:
                        entity.vx = 0
                        side = SIDE_LEFT
                        if _dx > 0: side = SIDE_RIGHT
                        dx = _dx = 0
                        entity.CollideWithLand(pos[0], pos[1], side)
                        break
            if dy != 0:            
                for pos in outline_y:
                    if self.level.GetLandAt(_new_x + pos[0],
                                            _new_y + pos[1]) != Level.LAND_NONE:
                        entity.vy *= -entity.elasticity
                        if abs(entity.vy) < 1: entity.vy = 0
                        side = SIDE_TOP
                        if _dy > 0: 
                            side = SIDE_BOTTOM
                            entity._physics_cache.last_bottom_collision_point = (_new_x + pos[0], _new_y + pos[1])
                        dy = _dy = 0
                        entity.CollideWithLand(pos[0], pos[1], side)
                        break
            # move to new position
            entity.x += _dx
            entity.y += _dy
            self.level.collidetector.DetectEntity(entity)
            if _dx == 0 and _dy == 0:
                return


class Scene:
    """
    The scene. (which means only the level. the level has everything)
    @ivar level: a Level object
    """
    
    def __init__(self, level):
        self.level = level
        self.objectMapById = self.level._all_entities
        self.physics_engine = PhysicsEngine(self.level)

    
    
    
