"""
This module defines resources, which are classes that uses to construct
other classes in the scene.

A resource shall be loaded from a file, and then be used to construct a certain
type of objects.
"""


import os
from m_Model import Outlines, Level
from m_View import LevelRep
from tools import Multifile
import pygame


class EntityResource:
    """
    Uses to construct Entity and EntityRep objects.
    
    This object contains the entity type, animations and etc.
    """

    class AnimPrototype:
        def __init__(self):
            self.frames = []
            self.delay = 0
            self.offset_x = 0
            self.offset_y = 0
    
    def __init__(self):
        self.animations = []
        self.states = [[0]] # list of active animation index lists
        self.type = 0
        self.mass = 0
        self.className = ""
        self.gravity = 0
        self.gravity_saturation = 0
        self.elasticity = 0
        self.outlines = Outlines()
        self.rect = pygame.Rect([0, 0, 0, 0])
        self.center = (0, 0)
    
    def Load(self, filename):
        """
        load this object with data from the given file name.
        """
        f = file(filename, 'r')
        _anim = None
        _transparent_pixel = (0, 0)
        while 1:
            line = f.readline()
            if len(line) == 0: break
            line = line.strip()
            if len(line) < 3 or line.find('#') == 0: continue
            kv = line.split(':')
            k, v = kv[0].strip(), kv[1]
            if k == 'class name':
                self.className = v.strip()
            elif k == 'transparent pixel':
                p = v.strip().rstrip(')').lstrip('(').split(',')
                _transparent_pixel = ( int(p[0].strip()), int(p[1].strip()) )
            elif k == 'type':
                self.type = int(v.strip())
            elif k == 'mass':
                self.mass = float(v.strip())
            elif k == 'elasticity':
                self.elasticity = float(v.strip())
            elif k == 'gravity':
                self.gravity= int(v.strip())
            elif k == 'gravity_saturation':
                self.gravity_saturation = int(v.strip())
            elif k == 'animation':
                v = v.strip()
                _anim = EntityResource.AnimPrototype()
                self.animations.append(_anim)
            elif k == 'animation.offset':
                p = v.strip().rstrip(')').lstrip('(').split(',')
                _anim.offset_x = int(p[0].strip())
                _anim.offset_y = int(p[1].strip())
            elif k == 'animation.delay':
                _anim.delay = int(v.strip())
            elif k == 'animation.image':
                path = os.path.join(os.path.dirname(filename), v.strip() )
                img = pygame.image.load(path).convert()
                img.set_colorkey(img.get_at(_transparent_pixel),
                                 pygame.locals.RLEACCEL)
                _anim.frames.append(img)
            elif k.find('outlines') != -1:
                ol = []
                while 1:
                    i, j = v.find('(') + 1, v.find(')')
                    if i == 0 or j - i < 2:
                        break
                    p = v[i:j].split(',')
                    v = v[j+1:]
                    ol.append( [ int(p[0].strip()), int(p[1].strip()) ] )
                k = k.split('.')[1].strip()
                if k == 'top': self.outlines.top = ol
                if k == 'bottom': self.outlines.bottom = ol
                if k == 'left': self.outlines.left = ol
                if k == 'right': self.outlines.right = ol
            elif k == 'rect':
                pp = v.strip().rstrip(')').lstrip('(').split(',')
                rect = [ int(pp[0].strip()), int(pp[1].strip()), \
                         int(pp[2].strip()), int(pp[3].strip()) ]
                self.rect = pygame.Rect(rect)
            elif k == 'center':
                p = v.strip().rstrip(')').lstrip('(').split(',')
                self.center = ( int(p[0].strip()), int(p[1].strip()) )
            elif k.find('states') != -1:
                self.states = []
                while 1:
                    i, j = v.find('(') + 1, v.find(')')
                    if i == 0 or j - i < 1:
                        break
                    pp = v[i:j].split(',')
                    state = []
                    for p in pp:
                        state.append(int (p.strip()))
                    self.states.append(state)
                    v = v[j+1:]



class LevelResource:
    """
    """
    
    def __init__(self):
        self._mfile = None # level multifile
        self.id = None
        self._is_init = False
        self.land_image = None
        self.sky_image = None
        self.rocks = {} # from file name to array of positions
        self.rock_images = {} # from file name to image
        
    def Load(self, filename):
        self._mfile = Multifile(filename)
        self.id = self._mfile.header.OID
        
    def _Init(self):
        f = self._mfile.GetFile('config.txt')
        while 1:
            line = f.readline()
            if len(line) == 0: break
            line = line.strip()
            if len(line) < 3 or line.find('#') == 0: continue
            kv = line.split(':')
            k, v = kv[0].strip(), kv[1]
            if k == 'land-image':
                fname = v.strip()
                _file = self._mfile.GetFile(fname)
                self.land_image = pygame.image.load(_file, fname)
                self.land_image.set_colorkey(self.land_image.get_at( (0, 0) ))
            elif k == 'sky-image':
                fname = v.strip()
                _file = self._mfile.GetFile(fname)
                self.sky_image = pygame.image.load(_file, fname).convert()
            elif k == 'rock':
                p, n = v.split(';')
                x, y = p.split(',')
                x = int( x.strip(' (') )
                y = int( y.strip(' )') )
                n = n.strip(' ')
                if self.rocks.has_key(n):
                    self.rocks[n].append( (x, y) )
                else:
                    self.rocks[n] = [ (x, y) ]
        for fname in self.rocks:
            _file = self._mfile.GetFile(fname)
            img = pygame.image.load(_file, fname).convert()
            img.set_colorkey(img.get_at((0, 0)))
            self.rock_images[fname] = img
        self.sky_image = pygame.transform.scale(\
                self.sky_image, self.land_image.get_size())
        self._is_init = True

    def CreateLevel(self):
        if not self._is_init: self._Init()
        land_map = pygame.surfarray.array_colorkey(self.land_image)
        w, h = self.land_image.get_width(), self.land_image.get_height()
        for fname in self.rocks:
            img = self.rock_images[fname]
            k = img.get_at((0, 0))
            for pos in self.rocks[fname]:
                for x in xrange(max(0, -pos[0]), min(w - pos[0], img.get_width())):
                    for y in xrange(max(0, -pos[1]), min(h - pos[1], img.get_height())):
                        if img.get_at((x, y)) != k:
                            x0 = pos[0] + x
                            y0 = pos[1] + y
                            land_map[x0][y0] = Level.LAND_ROCK
        level = Level(land_map, w, h)
        return level
        
    def CreateLevelRep(self):
        land_image = self.land_image.convert() # make a copy 
        for fname in self.rocks:
            img = self.rock_images[fname]
            for pos in self.rocks[fname]:
                land_image.blit(img, pos)
        return LevelRep(land_image=land_image, sky_image=self.sky_image)
    
    # added by zeffy
    def GetLevelSize(self):
        if self._mfile:
            return self._mfile.GetSize()
        else:
            return None
            



        
    
