"""
This module is the View part of the scene.

The objects in this module are responsible for knowing how the objects
in the model should be painted.
"""


import pygame
from m_Model import *
import time # for slowing down the thread


class EntityRep:
    """
    Entity Representation. Contains Animations for Entity objects.
    
    This object can tell what is the Entity's current image.
    
    @ivar type: an integer that is unique to a class of Entities.
                this value uses to connect between eitities with their 
                representation.
    """
    
    class Anim:
        def __init__(self):
            self.images = []
            self.offset_x = 0
            self.offset_y = 0
    
    def __init__(self, res):
        """
        Constructs a new EntityRep object.
        @param res: an EntityResource object.
        """
        self.type = res.type
        self.anims = []
        for a in res.animations:
            anim = EntityRep.Anim()
            self.anims.append(anim)
            anim.offset_x = a.offset_x
            anim.offset_y = a.offset_y
            anim.images = a.frames[:]
        


class LevelRep(ILevelObserver):
    """
    Level Representation. Responsible for drawing the land
    and sky (the background) of a level.
    """
    
    def __init__(self, land_image, sky_image):
        """
        Constructs a new LevelRep object.
        @param land_image: an image of the land.
        @param sky_image: an image of the sky.
        """
        self._land_image = land_image # .convert()
        self._sky_image = sky_image # .convert()
        self._image = self._sky_image.convert()
        self._image.blit(self._land_image, (0, 0) )
        
    def GetLevelImage(self):
        return self._image
        
    def NotifyLandRemoval(self, level, x, y, shape):
        """
        Callback which notifies about land removal.
        @param level: the Level object.
        """
        LAND_NONE, LAND_SAND = Level.LAND_NONE, Level.LAND_SAND
        w, h  = level.GetWidth() - 1, level.GetHeight() - 1
        land_map = level._land_map
        img_set_at = self._image.set_at
        img_get_at = self._sky_image.get_at
        for (y0, x0, x1) in shape.hlines:
            y0 += y
            if y0 < 0 or y0 > h: continue
            x0 = max(0, x + x0)
            x1 = min(w, x + x1)
            while x0 < x1:
                if land_map[x0][y0] == LAND_NONE:
                    img_set_at((x0, y0), img_get_at((x0, y0)))
                x0 += 1
            time.sleep(0)



class SceneRep:
    """
    scene representation.
    """
    
    def __init__(self):
        self.level_rep = None
        self.entity_rep_map = {}
        
    def CreateLevelRep(self, level, res):
        self.level_rep = res.CreateLevelRep()
        level.AttachObserver(self.level_rep)

    def CreateEntityRep(self, res):
        self.entity_rep_map[res.type] = EntityRep(res)
        
    def GetLevelImage(self):
        return self.level_rep.GetLevelImage()
        
    def GetEntityRep(self, entity):
        if self.entity_rep_map.has_key(entity.type):
            return self.entity_rep_map[entity.type]
        else:
            return None
        


class Camera:
    
    def __init__(self, scene, scene_rep, output_surface):
        """
        The position of the camera relates to the center of
        the camera's view pots.
        """
        self.scene = scene
        self.scene_rep = scene_rep
        self.output_surface = output_surface
        self.x = 0
        self.y = 0
        self.x_offset = 0
        self.y_offset = 0
        self.w, self.h = output_surface.get_size()
        self.lev_w = scene.level.GetWidth()
        self.lev_h = scene.level.GetHeight()
        
    def SetPos(self, x, y):
        self.x, self.y = x, y
        self.x_offset = int( self.x - self.w / 2 )
        self.y_offset = int( self.y - self.h / 2 )
        if self.x_offset < 0: self.x_offset = 0
        if self.x_offset + self.w > self.lev_w:
            self.x_offset = self.lev_w - self.w
        if self.y_offset < 0: self.y_offset = 0
        if self.y_offset + self.h > self.lev_h:
            self.y_offset = self.lev_h - self.h
        
    def Paint(self):
        self.output_surface.blit(
             self.scene_rep.GetLevelImage(),
             (0, 0),
             (self.x_offset, self.y_offset, self.w, self.h))
        for entity in self.scene.level.GetEntities():
            rep = self.scene_rep.GetEntityRep(entity)
            for i in entity.active_anim_indexes:
                anim = rep.anims[i]
                img = anim.images[entity.anims[i].frame_index]
                self.output_surface.blit(
                        img,
                        (int(entity.x) - self.x_offset + anim.offset_x,
                         int(entity.y) - self.y_offset + anim.offset_y))
                     
    def PaintObj(self, x, y, surface):
        self.output_surface.blit(
                surface,
                (int(x) - self.x_offset, int(y) - self.y_offset))
        
 
    
